package net.rmi.rjs.jobs;

import gui.ClosableJFrame;
import gui.In;
import j2d.Images;
import j2d.ShortImageBean;

import java.awt.*;
import java.io.Serializable;
import java.rmi.RemoteException;

public class Fractals {
    private final FractalLogic fractalLogic = new FractalLogic();

    public Fractals() {
    }
    private  ComputeJobs getComputeJobs(Dimension d) {
        ComputeJobs cjs = new ComputeJobs();
        cjs.add(getComputeJob(d.width, d.height));
        cjs.add(getComputeJob(d.width, d.height));
        return cjs;
    }

    public static void main(String args[]) {
        //testMandelbrot();
        Fractals f = new Fractals();
        f.testComputeJobs();
    }


    public void testComputeJobs() {
        final Dimension jobDimension = new Dimension(200, 400);

        fractalLogic.getAns().add(getComputeJobs(jobDimension));
        ClosableJFrame cjf = new MultiThreadedFrame(jobDimension);
        cjf.setSize(new Dimension(400, 400));
        cjf.show();
    }

    public  ComputeJob getComputeJob(final int w, final int h) {
        return new ComputeJob() {
            public Serializable getAnswer() {
                ShortImageBean sib = new ShortImageBean(w, h);
                fractalLogic.mandelbrot(sib.getR(), sib.getG(), sib.getB());
                return sib;
            }
        };
    }

    private void testComputeJob() {
        ClosableJFrame cjf = new ClosableJFrame("mandlebrot") {
            public void paint(Graphics g) {
                try {
                    Dimension d = getSize();
                    ComputeJob cj = getComputeJob(
                            d.width, d.height);
                    Serializable s = cj.getAnswer();
                    if (s instanceof ShortImageBean) {
                        ShortImageBean sib = (ShortImageBean) s;
                        g.drawImage(sib.getImage(), 0, 0, this);
                    }
                } catch (RemoteException e) {
                    In.message(e);

                }
            }

        };
        cjf.setSize(400, 400);
        cjf.show();
    }


    public FractalLogic getFractalLogic() {
        return fractalLogic;
    }

    private class MultiThreadedFrame extends ClosableJFrame {
        private final Dimension jobDimension;


        Images imgs = new Images();
        Answers ans = fractalLogic.getAns();
        Serializable sa[] = ans.getAnswers();

        public MultiThreadedFrame(Dimension jobDimension) {
            super("mandlebrot");
            this.jobDimension = jobDimension;
            
        }

        public void paint(Graphics g) {


            System.out.println("got:"+sa.length+" answers.");
            for (int i = 0; i < sa.length; i++) {

                Serializable s = sa[i];
                if (s instanceof ShortImageBean) {
                    ShortImageBean sib = (ShortImageBean) s;
                    imgs.add(sib.getImage());
                    g.drawImage(imgs.getImage(jobDimension, 1, 2), 0, 0, this);
                }
            }

        }
    }
}